#include <oskit/sproc.h>
#include <oskit/debug.h>
#include <oskit/c/malloc.h>
#include <oskit/threads/pthread.h>
#include <oskit/threads/cpuinherit.h>
#include <oskit/version.h>
#include <oskit/x86/proc_reg.h>
#include <oskit/x86/trap.h>
#include <oskit/page.h>
#include <stdio.h>
#include <string.h>

#include <kernel/process.h>
#include <kernel/process_defs.h>
#include <kernel/syscall_defs.h>
#include <kernel/message_defs.h>

#define USER_STACK_SIZE (16*1024)

/*
  The Idle Thread (TM)
 */

pthread_t idlethread;

void idle_function () 
{
	while (1);
}

/*
  Process list
*/

#define MAX_PROCESS_COUNT 50
mth_process process_pool[MAX_PROCESS_COUNT];
int process_pool_used[MAX_PROCESS_COUNT];
mth_process* process_table[MAX_PROCESS_COUNT];
pthread_mutex_t table_lock;
int process_count = 1;
int process_number = 1;

/*
  Inter-Process Communication
 */

#define MAX_MESSAGE_COUNT 50
mth_message message_pool[MAX_MESSAGE_COUNT];
int message_pool_used[MAX_MESSAGE_COUNT];
mth_message* message_table[MAX_PROCESS_COUNT];
pthread_mutex_t message_table_lock;
int message_count = 0;

/*
  Thread scheduling parameters
 */
struct sched_param scheduling_parameter;
struct sched_param *scheduling_parameter_p;

void mth_process_init () 
{
	int i = 0;
	int result = 0;
	
	for (i = 0; i < MAX_PROCESS_COUNT; i++) {
		process_pool_used[i] = 0;	
		process_table[i] = 0;
	}

	result = pthread_create(&idlethread, NULL, (void*(*)(void*))idle_function, NULL);
	if (result != 0) {
		printf ("mth_process_init: error creating idle thread!\n");
		return;
	}
	methos_sched_setidlethread ((int)idlethread);
	
	scheduling_parameter_p = &scheduling_parameter;
	pthread_mutex_init (&table_lock, 0);
	pthread_mutex_init (&message_table_lock, 0);
}

mth_process *
allocate_process_structure () 
{
	int i;
	mth_process *proc = NULL;

	pthread_mutex_lock (&table_lock);
	for (i = 0; i < MAX_PROCESS_COUNT; i++) {
		if (process_pool_used[i] == 0) {
			proc = &process_pool[i];
			process_pool_used[i] = 1;
			break;
		}
	}
	pthread_mutex_unlock (&table_lock);
	return proc;
}

void 
deallocate_process_structure (mth_process* process) 
{
	int i;

	pthread_mutex_lock (&table_lock);		
	for (i = 0; i < MAX_PROCESS_COUNT; i++) {
		if (&process_pool[i] == process) {
			process->procnum = -1;
			strcpy (process->name, "no process");
			process_pool_used[i] = 0;
		}
	}
	pthread_mutex_unlock (&table_lock);
}

mth_process *
table_get_process_by_num (int procnum)
{
	int i;
	mth_process* proc = 0;

	pthread_mutex_lock (&table_lock);
	for (i = 1; i < MAX_PROCESS_COUNT; i++) {
		if (process_table[i] != 0) {
			if (process_table[i]->procnum == procnum)
				proc = process_table[i];
		}
	}
	pthread_mutex_unlock (&table_lock);
	return proc;
}

mth_process *
table_get_process_by_name (const char* name)
{
	mth_process* proc = NULL;
	int i;

	pthread_mutex_lock (&table_lock);
	for (i = 1; i < MAX_PROCESS_COUNT; i++) {
		if (process_table[i] != 0) {
			if (strcmp (process_table[i]->name, name) == 0)
				proc =  process_table[i];
		}
	}
	pthread_mutex_unlock (&table_lock);
	return proc;
}

void 
table_add_process (mth_process* proc)
{
	int i;
	pthread_mutex_lock (&table_lock);
	for (i = 1; i < MAX_PROCESS_COUNT; i++) {
		if (process_table[i] == 0) {
			process_table[i] = proc;
			pthread_mutex_unlock (&table_lock);
			while (table_get_process_by_num (process_number) != 0) 
				process_number = (process_number + 1) % 25000;
			pthread_mutex_lock (&table_lock);
			proc->procnum = process_number;
			process_count++;
			break;
		}
	}
	pthread_mutex_unlock (&table_lock);
}

void
table_remove_process (int procnum)
{
	int i = 0;

	pthread_mutex_lock (&table_lock);
	for (i = 1; i < process_count; i++) {
		if (process_table[i]->procnum == procnum) {
			process_table[i] = 0;
		}
	}

	process_count--;
	pthread_mutex_unlock (&table_lock);	
}

/*
  Process data structure for the Simple Process Library
*/
static int
mth_trap_handler(struct oskit_sproc_thread *sthread, int signo, int code,
        struct trap_state *ts)
{
	printf("\n\nMethOS Error: thread = %d, process = %p, signo = %d, "
           "code = %d, frame = %p\n",
           (int)pthread_self(), sthread->st_process, signo, code, ts);
	return 1;
}

extern struct oskit_sproc_sysent methos_syscall_table[]; 
static struct oskit_sproc_desc process_desc = {
    METHOS_SYSCALL_COUNT,
    methos_syscall_table,
    mth_trap_handler
};

static void mth_execute_process(void *);

static void
mth_load_process(struct oskit_sproc *sproc, exec_info_t *exec_info, const char *elf){
	oskit_addr_t heap;
	oskit_error_t error;

	if (oskit_sproc_create (&process_desc, PROCESS_SIZE, sproc)) {
		panic("oskit_sproc_create failed\n");
	}

	error = oskit_sproc_load_elf (sproc, elf, exec_info);
	if (error) {
		printf ("Error loading program '%s'. Maybe it is busy?\n", elf);
		sproc->sp_flags = -1;
		return;
	}

	heap = HEAP_START_ADDR;
	error = oskit_uvm_mmap(sproc->sp_vm, &heap, HEAP_SIZE, PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANON|MAP_FIXED, 0, 0);
	if (error) {
		panic("oskit_uvm_mmap failed (heap)\n");
	}
	assert(heap == HEAP_START_ADDR);
}

struct lightweight_process_arg {
	struct oskit_sproc    *sproc;
	oskit_addr_t  entry;
};

static void
mth_lightweight_process (struct lightweight_process_arg *arg)
{
	oskit_addr_t stkaddr;
	struct oskit_sproc_stack stk;
	stkaddr = 0;

	if (oskit_sproc_stack_alloc(arg->sproc, &stkaddr, USER_STACK_SIZE, PAGE_SIZE, &stk)) {
		panic("oskit_sproc_alloc_stack failed\n");
	}
	
	oskit_sproc_switch (arg->sproc, arg->entry, &stk);
}


static void
mth_execute_process(void *arg)
{
	exec_info_t exec_info;
	const char *filename = (char *)arg;
	struct lightweight_process_arg lwp_arg;
	int result = 0;
	mth_process* proc = table_get_process_by_name (filename);

	mth_load_process (&proc->sproc, &exec_info, filename);

	if (proc->sproc.sp_flags == -1) {
		oskit_sproc_destroy(&proc->sproc);
		table_remove_process (proc->procnum);
		deallocate_process_structure (proc);
		return;
	}
		
	lwp_arg.sproc = &proc->sproc;
	lwp_arg.entry = exec_info.entry;

	result = pthread_create(&proc->process_thread, NULL, (void*(*)(void*))mth_lightweight_process, &lwp_arg);
	if (result != 0) {
		printf ("mth_execute_process: error creating thread!\n");
		return;
	}

  	proc->running = 1;

	pthread_join(proc->process_thread, NULL);
	
	proc->running = 0;

	oskit_sproc_destroy(&proc->sproc);
	table_remove_process (proc->procnum);
	deallocate_process_structure (proc);
}

int
mth_process_create () 
{
	mth_process* proc = allocate_process_structure ();
	if (proc == NULL) {
		printf ("mth_process_create: Failed to allocate process structure.\n");
		return 0;
	}
	proc->running = 0;
	proc->type = TTYPE_INTERACTIVE;

	table_add_process (proc);
	return proc->procnum;
}

int mth_process_loadelf (int procnum, const char* filename) 
{
	int success = 0;
	mth_process* proc = table_get_process_by_num (procnum);

	if (filename == 0) {
		printf ("mth_process_loadelf: null filename passed!\n");
		return 1;
	}

	strncpy (proc->name, filename, MAX_PROCESS_NAME);

	success = pthread_create(&proc->control_thread, NULL, (void *(*)(void*))mth_execute_process, (char*)proc->name);
	if (success != 0) {
		printf ("mth_process_loadelf: Error creating control thread!\n");
		return 1;
	}
	
	return 0;
}

void mth_process_join (int procnum) 
{
	mth_process* proc = table_get_process_by_num (procnum);
	if (proc != 0)
		pthread_join(proc->control_thread, NULL);
}

void
mth_process_kill (int procnum)
{
	mth_process* proc = table_get_process_by_num (procnum);
	if (proc != 0)
		pthread_cancel(proc->process_thread);
}


int 
mth_process_self () 
{
	mth_process* proc = NULL;
	int i;

	for (i = 0; i < MAX_PROCESS_COUNT; i++) {
		if (process_table[i] != 0) {
			if ((int)process_table[i]->process_thread == (int)pthread_self ())
				proc =  process_table[i];
		}
	}

	return proc->procnum;
}

int
mth_process_number (const char* name)
{
	mth_process* proc = table_get_process_by_name (name);
	if (proc == 0)
		return 0;
	else
		return proc->procnum;
}

void
mth_process_settype (int procnum, int newtype)
{
	mth_process* proc = table_get_process_by_num (procnum);	
	if (newtype != TTYPE_INTERACTIVE && newtype != TTYPE_BACKGROUND)
		return;

	if (proc == 0)
		return;

	proc->type = newtype;

	methos_sched_settype ((int)proc->process_thread, newtype);
}

int
mth_process_gettype (int procnum) 
{
	mth_process* proc = table_get_process_by_num (procnum);	
	return proc->type;	
}

void
mth_process_list () 
{
	int i;
	int pt, ct;
	mth_process* proc;
	
	printf ("Num\tNimi\t\tK-L6im\tL6im\tTyyp\tAjahulk\tCPU %%\tCPU Aega\n");

	printf ("0\tvaba\t\tn/a\tn/a\tn/a\tn/a");
	printf ("\t%d%%", methos_sched_cpupercent (3));
	printf ("\t%d", methos_sched_cputime (3));
	printf ("\n");

	pthread_mutex_lock (&table_lock);
	for (i = 1; i < process_count; i++) {
		proc = process_table[i];
		ct = (int)proc->control_thread;
		pt = (int)proc->process_thread;	
		if (proc->running == 1) {
			printf ("%d", proc->procnum);
			printf ("\t%s", proc->name);
			if (strlen (proc->name) < 8)
				printf ("\t");
			printf ("\t%d", ct);
			printf ("\t%d", pt);
			printf ("\t%s", (methos_sched_type (pt) == 0 ? "INTER" : "TAUST"));
			printf ("\t%d", methos_sched_timeslice (pt));
			printf ("\t%d%%", methos_sched_cpupercent (pt) + methos_sched_cpupercent (ct));
			printf ("\t%d", (methos_sched_cputime (pt) + methos_sched_cputime (ct)) * 10);
			printf ("\n");
		}
	}
	pthread_mutex_unlock (&table_lock);
}

void mth_message_init ()
{
	int i = 0;
	
	for (i = 0; i < MAX_MESSAGE_COUNT; i++) {
		message_pool_used[i] = 0;	
		message_table[i] = 0;
	}
}

mth_message *
allocate_message_structure () 
{
	int i;
	mth_message *msg = NULL;

	pthread_mutex_lock (&message_table_lock);
	for (i = 0; i < MAX_MESSAGE_COUNT; i++) {
		if (message_pool_used[i] == 0) {
			msg = &message_pool[i];
			strcpy (msg->data, "");
			msg->size = 0;
			msg->source = 0;
			msg->destination = 0;
			message_pool_used[i] = 1;
			break;
		}
	}
	pthread_mutex_unlock (&message_table_lock);
	return msg;
}

void 
deallocate_message_structure (mth_message* message) 
{
	int i;

	pthread_mutex_lock (&message_table_lock);
	for (i = 0; i < MAX_MESSAGE_COUNT; i++) {
		if (&message_pool[i] == message) {
			message_pool_used[i] = 0;
		}
	}
	pthread_mutex_unlock (&message_table_lock);
}

int mth_message_send (int dest, int size, char *data) 
{
	mth_message* msg = 0;
	mth_process* destproc = 0;

	if (message_count >= MAX_MESSAGE_COUNT) {
		printf ("mth_message_send: Internal queue full!\n");
		return 1;
	}

	destproc = table_get_process_by_num (dest);
	if (destproc == 0) {
		return 1;
	}
	
	msg = allocate_message_structure ();

	pthread_mutex_lock (&message_table_lock);
	msg->source = mth_process_self ();
	msg->destination = destproc->procnum;
	msg->size = size;
	strcpy (msg->data, data);
	
	message_table[message_count] = msg;
	message_count++;

	pthread_mutex_unlock (&message_table_lock);	
	return 0;
}

int mth_message_receive (int *sender, int *size, char *data)
{
	int i = 0;
	int j = 0;
	mth_process* senderproc = 0;
	int me = mth_process_self ();

	pthread_mutex_lock (&message_table_lock);
	
	for (i = 0; i < message_count; i++) {
		if (message_table[i] != 0) {
			if (message_table[i]->destination == me) {
				if (sender != 0) {
					senderproc = table_get_process_by_num (message_table[i]->source);
					if (senderproc)
						*sender = senderproc->procnum;
					else
						*sender = 0;
				}
				if (size != 0) {
					*size = message_table[i]->size;
				}
				if (data != 0) {
					strcpy (data, message_table[i]->data);
				}
				else {
					pthread_mutex_unlock (&message_table_lock);	
					return 1;
				}

				pthread_mutex_unlock (&message_table_lock);	

				deallocate_message_structure (message_table[i]);

				pthread_mutex_lock (&message_table_lock);	

				message_count--;
				for (j = i; j < message_count; j++) {
					message_table[j] = message_table[j + 1];
				}

				pthread_mutex_unlock (&message_table_lock);	
	
				return 0;
			}
		}
	}
	pthread_mutex_unlock (&message_table_lock);	
	return 1;
}

